import asyncio
import collections
import itertools
import time

from .protocol import Protocol, build_message, Request, Payload


class ConnectionProtocol(asyncio.BufferedProtocol):
    """
    start -> CM [-> GB [-> BU?]]* [-> ER?] -> CL -> end
    """

    def __init__(self, on_connection=None, loop=None, **kwargs):
        super().__init__()
        self._on_connection = on_connection
        self.transport = None
        self.connection = None
        self._loop = loop or asyncio.get_running_loop()
        self._connection_lost = None
        self._paused = False
        self._drain_waiter = None
        self.protocol = Protocol(self.message_received, **kwargs)

    def connection_made(self, transport):
        self.transport = transport
        self.connection = Connection(self, self.transport, loop=self._loop)

        if self._on_connection is not None:
            self._on_connection(self.connection)

    def get_buffer(self, sizehint):
        return self.protocol.get_buffer()

    def message_received(self, kind, args):
        self.connection._append_msg(build_message(kind, args))

    def buffer_updated(self, nbytes):
        self.protocol.buffer_updated(nbytes)

    def eof_received(self):
        self.connection._set_exception(ConnectionResetError())

    def connection_lost(self, exc=None):
        if exc is None:
            exc = ConnectionResetError("Connection closed")
        self.connection._set_exception(exc)
        self._connection_lost = exc

        if self._paused:
            waiter = self._drain_waiter
            if waiter is not None:
                self._drain_waiter = None
                if not waiter.done():
                    waiter.set_exception(exc)

    def pause_writing(self):
        self._paused = True

    def resume_writing(self):
        self._paused = False

        waiter = self._drain_waiter
        if waiter is not None:
            self._drain_waiter = None
            if not waiter.done():
                waiter.set_result(None)

    async def write(self, msg):
        parts = msg.serialize()
        if len(parts) > 1:
            self.transport.writelines(parts)
        else:
            self.transport.write(parts[0])
        if self.transport.is_closing():
            await asyncio.sleep(0, loop=self._loop)
        elif self._paused and not self._connection_lost:
            self._drain_waiter = self._loop.create_future()
            await self._drain_waiter


class Connection(object):
    """A connection between two endpoints.

    Use ``connect`` to create a connection.
    """

    def __init__(self, protocol, transport, loop):
        self._protocol = protocol
        self._transport = transport
        self._loop = loop

        self._id_iter = itertools.count()
        self._active_reqs = {}
        self._queue = collections.deque()
        self._yield_cycler = itertools.cycle(range(50))
        self._waiter = None
        self._exception = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, typ, value, traceback):
        await self.close()
        if isinstance(value, ConnectionResetError):
            return True

    async def _maybe_yield(self):
        if not next(self._yield_cycler):
            await asyncio.sleep(0, loop=self._loop)

    async def request(self, route, metadata=None, body=None):
        """Send a request message and wait for a response"""
        if self._exception is not None:
            raise self._exception

        msg_id = next(self._id_iter)
        msg = Request(msg_id, route, metadata=metadata, body=body)
        reply = self._active_reqs[msg_id] = self._loop.create_future()
        await self._protocol.write(msg)
        return await reply

    async def send(self, msg):
        if self._exception is not None:
            raise self._exception
        await self._protocol.write(msg)
        await self._maybe_yield()

    async def __aiter__(self):
        try:
            while True:
                yield await self.recv()
        except ConnectionResetError:
            await self.close()

    async def recv(self):
        """Wait for the next message"""
        if self._exception is not None:
            raise self._exception

        if not self._queue:
            if self._waiter is not None:
                raise RuntimeError(
                    "Connection.recv may only be called by one coroutine at a time"
                )
            self._waiter = self._loop.create_future()
            try:
                await self._waiter
            finally:
                self._waiter = None

        return self._queue.popleft()

    def _close(self):
        if self._transport is not None:
            transport = self._transport
            self._transport = None
            return transport.close()

    async def close(self):
        """Close the connection and release all resources.

        It is invalid to use this connection after closing.

        This method is idempotent.
        """
        self._close()
        try:
            futs = self._active_reqs.values()
            await asyncio.gather(*futs, return_exceptions=True)
        except asyncio.CancelledError:
            pass

    def _append_msg(self, msg):
        if isinstance(msg, Request):
            self._queue.append(msg)

            waiter = self._waiter
            if waiter is not None:
                self._waiter = None
                waiter.set_result(False)
        elif isinstance(msg, Payload):
            message = self._active_reqs.pop(msg.id, None)
            if message is not None and not message.done():
                message.set_result(msg)
        else:
            self._set_exception(RuntimeError("Invalid message %s" % msg))

    def _set_exception(self, exc):
        if self._exception:
            return

        self._exception = exc

        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_exception(exc)

        self._close()


async def connect(addr, *, loop=None, timeout=0, **kwargs):
    """Create a new connection.

    Parameters
    ----------
    addr : tuple or str
        The address to connect to.
    loop : AbstractEventLoop, optional
        The event loop to use.
    timeout : float, optional
        Timeout for initial connection to the server.
    **kwargs
        All remaining arguments are forwarded to ``loop.create_connection``.

    Returns
    -------
    connection : Connection
    """
    if loop is None:
        loop = asyncio.get_event_loop()

    if timeout is None:
        timeout = float("inf")

    def factory():
        return ConnectionProtocol(loop=loop)

    if isinstance(addr, tuple):
        connect = loop.create_connection
        args = (factory,) + addr
        connect_errors = (ConnectionRefusedError, OSError)
    elif isinstance(addr, str):
        connect = loop.create_unix_connection
        args = (factory, addr)
        connect_errors = FileNotFoundError
    else:
        raise ValueError("Unknown address type: %s" % addr)

    retry_interval = 0.5
    start_time = time.monotonic()
    while True:
        try:
            _, p = await connect(*args, **kwargs)
            break
        except connect_errors:
            if (time.monotonic() - start_time) > timeout:
                raise
            await asyncio.sleep(retry_interval)
            retry_interval = min(30, 1.5 * retry_interval)

    return p.connection



import struct

from . import _lib


Protocol = _lib.Protocol


FLAG_METADATA = 1 << 7
FLAG_BODY = 1 << 6
FLAG_FRAMES = 1 << 5
FLAG_NEXT = 1 << 4
FLAG_COMPLETE = 1 << 3


u32_struct = struct.Struct("<L")
pack_u32 = u32_struct.pack_into

u16_struct = struct.Struct("<H")
pack_u16 = u16_struct.pack_into


def _write(
    kind,
    id=None,
    uint32=None,
    route=None,
    metadata=None,
    body=None,
    is_next=False,
    is_complete=False,
):
    nframes = route_length = None
    length = 2
    if id is not None:
        length += 4
    if uint32 is not None:
        length += 4
    flags = 0
    if route is not None:
        route_length = len(route)
        length += 2 + route_length
    if metadata is not None:
        flags |= FLAG_METADATA
        length += 4
    if body is not None:
        flags |= FLAG_BODY
        if isinstance(body, (list, tuple)):
            flags |= FLAG_FRAMES
            nframes = len(body)
            length += 2 + 4 * nframes
        else:
            length += 4
    if is_next:
        flags |= FLAG_NEXT
    if is_complete:
        flags |= FLAG_COMPLETE

    header = bytearray(length)
    header[0] = kind
    header[1] = flags
    chunks = [header]
    offset = 2
    if id is not None:
        pack_u32(header, offset, id)
        offset += 4
    if route is not None:
        pack_u16(header, offset, route_length)
        offset += 2
    if uint32 is not None:
        pack_u32(header, offset, uint32)
        offset += 4
    if metadata is not None:
        pack_u32(header, offset, len(metadata))
        offset += 4
        chunks.append(metadata)
    if body is not None:
        if isinstance(body, (list, tuple)):
            pack_u16(header, offset, nframes)
            offset += 2
            for frame in body:
                frame_length = len(frame)
                pack_u32(header, offset, frame_length)
                offset += 4
                if frame_length > 0:
                    chunks.append(frame)
        else:
            body_length = len(body)
            pack_u32(header, offset, body_length)
            offset += 4
            if body_length > 0:
                chunks.append(body)
    if route is not None:
        header[offset:] = route
    return chunks


class Setup(object):
    __slots__ = ("heartbeat", "metadata")

    def __init__(self, heartbeat, metadata=None):
        self.heartbeat = heartbeat
        self.metadata = metadata

    def serialize(self):
        return _write(_lib.KIND_SETUP, uint32=self.heartbeat, metadata=self.metadata,)


class SetupResponse(object):
    __slots__ = ("heartbeat", "metadata")

    def __init__(self, heartbeat, metadata=None):
        self.heartbeat = heartbeat
        self.metadata = metadata

    def serialize(self):
        return _write(
            _lib.KIND_SETUP_RESPONSE, uint32=self.heartbeat, metadata=self.metadata,
        )


class Heartbeat(object):
    __slots__ = ()

    def serialize(self):
        return [_lib.KIND_HEARTBEAT.to_bytes(1, "big", signed=True)]


class Error(object):
    __slots__ = ("id", "code", "metadata", "body")

    def __init__(self, id, code, metadata=None, body=None):
        self.id = id
        self.code = code
        self.metadata = metadata
        self.body = body

    def serialize(self):
        return _write(
            _lib.KIND_ERROR,
            id=self.id,
            uint32=self.code,
            metadata=self.metadata,
            body=self.body,
        )


class Cancel(object):
    __slots__ = ("id",)

    def __init__(self, id):
        self.id = id

    def serialize(self):
        return _write(_lib.KIND_CANCEL, id=self.id)


class IncrementWindow(object):
    __slots__ = ("id", "window")

    def __init__(self, id, window):
        self.id = id
        self.window = window

    def serialize(self):
        return _write(_lib.KIND_INCREMENT_WINDOW, id=self.id, uint32=self.window)


class Request(object):
    __slots__ = ("id", "route", "metadata", "body")

    def __init__(self, id, route, metadata=None, body=None):
        self.id = id
        self.route = route
        self.metadata = metadata
        self.body = body

    def serialize(self):
        return _write(
            _lib.KIND_REQUEST,
            id=self.id,
            route=self.route,
            metadata=self.metadata,
            body=self.body,
        )


class Stream(object):
    __slots__ = ("id", "route", "window", "metadata", "body")

    def __init__(self, id, window, route, metadata=None, body=None):
        self.id = id
        self.window = window
        self.route = route
        self.metadata = metadata
        self.body = body

    def serialize(self):
        return _write(
            _lib.KIND_REQUEST_STREAM,
            id=self.id,
            uint32=self.window,
            route=self.route,
            metadata=self.metadata,
            body=self.body,
        )


class Channel(object):
    __slots__ = ("id", "route", "window", "metadata", "body")

    def __init__(self, id, window, route, metadata=None, body=None):
        self.id = id
        self.window = window
        self.route = route
        self.metadata = metadata
        self.body = body

    def serialize(self):
        return _write(
            _lib.KIND_REQUEST_CHANNEL,
            id=self.id,
            uint32=self.window,
            route=self.route,
            metadata=self.metadata,
            body=self.body,
        )


class Payload(object):
    __slots__ = ("id", "metadata", "body", "is_next", "is_complete")

    def __init__(self, id, metadata=None, body=None, is_next=False, is_complete=False):
        self.id = id
        self.metadata = metadata
        self.body = body
        self.is_next = is_next
        self.is_complete = is_complete

    def serialize(self):
        return _write(
            _lib.KIND_PAYLOAD,
            id=self.id,
            metadata=self.metadata,
            body=self.body,
            is_next=self.is_next,
            is_complete=self.is_complete,
        )


dispatch_table = {
    _lib.KIND_SETUP: Setup,
    _lib.KIND_SETUP_RESPONSE: SetupResponse,
    _lib.KIND_HEARTBEAT: Heartbeat,
    _lib.KIND_ERROR: Error,
    _lib.KIND_CANCEL: Cancel,
    _lib.KIND_INCREMENT_WINDOW: IncrementWindow,
    _lib.KIND_REQUEST: Request,
    _lib.KIND_STREAM: Stream,
    _lib.KIND_CHANNEL: Channel,
    _lib.KIND_PAYLOAD: Payload,
}


def build_message(kind, args):
    return dispatch_table[kind](*args)



import os

from setuptools import setup
from setuptools.extension import Extension

ext_modules = [Extension("ery._lib", ["ery/_lib.c"])]

setup(
    name="ery",
    version="0.0.1",
    maintainer="Jim Crist-Harif",
    maintainer_email="jcristharif@gmail.com",
    license="BSD",
    packages=["ery"],
    ext_modules=ext_modules,
    long_description=(
        open("README.rst").read() if os.path.exists("README.rst") else ""
    ),
    zip_safe=False,
)



import asyncio
import os
import socket
import weakref

from .aioprotocol import ConnectionProtocol


class Listener(object):
    """Base class for all listeners"""

    def __init__(self, server, ssl=None, **kwargs):
        self.server = server
        self._ssl = ssl
        self._extra_kwargs = kwargs
        self._handle = None

    async def start(self):
        pass

    async def stop(self):
        if self._handle is None:
            return
        self._handle.close()
        await self._handle.wait_closed()


class TCPListener(Listener):
    def __init__(self, server, port, host=None, **kwargs):
        super().__init__(server, **kwargs)
        if host is None:
            host = "0.0.0.0"

        scheme = "tls" if self._ssl else "tcp"
        self.name = f"{scheme}://{host}:{port}"
        self._host = host
        self._port = port

    async def start(self) -> None:
        await super().start()
        loop = asyncio.get_event_loop()
        self._handle = await loop.create_server(
            self.server._protocol_factory,
            host=self._host,
            port=self._port,
            ssl=self._ssl,
            **self._extra_kwargs,
        )


class UnixListener(Listener):
    def __init__(self, server, path, **kwargs):
        super().__init__(server, **kwargs)
        path = os.path.abspath(path)
        scheme = "unix+tls" if self._ssl else "unix"
        self.name = f"{scheme}:{path}"
        self._path = path

    async def start(self) -> None:
        await super().start()
        loop = asyncio.get_event_loop()
        self._handle = await loop.create_unix_server(
            self.server._protocol_factory,
            path=self._path,
            ssl=self._ssl,
            **self._extra_kwargs,
        )


class SocketListener(Listener):
    def __init__(self, server, sock, **kwargs):
        super().__init__(server, **kwargs)

        if hasattr(socket, "AF_UNIX") and sock.family == socket.AF_UNIX:
            path = os.path.abspath(sock.getsockname())
            scheme = "unix+tls" if self._ssl else "unix"
            self.name = f"{scheme}:{path}"
        else:
            host, port = sock.getsockname()[:2]
            scheme = "tls" if self._ssl else "tcp"
            self.name = f"{scheme}://{host}:{port}"
        self._sock = sock

    async def start(self) -> None:
        await super().start()
        loop = asyncio.get_event_loop()
        self._handle = await loop.create_server(
            self.server._protocol_factory,
            sock=self._sock,
            ssl=self._ssl,
            **self._extra_kwargs,
        )


class Server(object):
    def __init__(self, handler):
        self.handler = handler
        self.listeners = []
        self._handlers = weakref.WeakSet()

    def _protocol_factory(self):
        return ConnectionProtocol(self._on_connection)

    def _on_connection(self, connection):
        task = asyncio.ensure_future(self.handler(connection))
        self._handlers.add(task)

    def add_tcp_listener(self, port, *, host=None, ssl=None, **kwargs):
        """Add a new TCP listener.

        Parameters
        ----------
        port : int
            The port to listen on.
        host : str or list[str], optional
            The host (or hosts) to listen on. Default is all interfaces.
        ssl : SSLContext, optional
            If provided, TLS will be used over accepted connections.
        **kwargs : optional
            Additional parameters to forward to ``asyncio.EventLoop.create_server``.
        """
        self.listeners.append(TCPListener(self, port, host=host, ssl=ssl, **kwargs))

    def add_unix_listener(self, path, *, ssl=None, **kwargs):
        """Add a new Unix listener.

        Parameters
        ----------
        path : str
            The path of the unix domain socket to listen on.
        ssl : SSLContext, optional
            If provided, TLS will be used over accepted connections.
        **kwargs : optional
            Additional parameters to forward to ``asyncio.EventLoop.create_unix_server``.
        """
        self.listeners.append(UnixListener(self, path, ssl=ssl, **kwargs))

    def add_socket_listener(self, sock, *, ssl=None, **kwargs):
        """Add a new listener on an already created socket.

        Parameters
        ----------
        sock : socket.socket
            An already created socket object.
        ssl : SSLContext, optional
            If provided, TLS will be used over accepted connections.
        **kwargs : optional
            Additional parameters to forward to ``asyncio.EventLoop.create_server``.
        """
        self.listeners.append(SocketListener(self, sock, ssl=ssl, **kwargs))

    async def start(self):
        await asyncio.gather(*(listener.start() for listener in self.listeners))

    async def stop(self):
        await asyncio.gather(*(listener.stop() for listener in self.listeners))
        await asyncio.gather(*self._handlers, return_exceptions=True)
